/***
*
*	Copyright (c) 1996-2001, Valve LLC. All rights reserved.
*	
*	This product contains software technology licensed from Id 
*	Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc. 
*	All Rights Reserved.
*
*   Use, distribution, and modification of this source code and/or resulting
*   object code is restricted to non-commercial enhancements to products from
*   Valve LLC.  All other use, distribution, or modification is prohibited
*   without written permission from Valve LLC.
*
****/
//=========================================================
// Monster Maker - this is an entity that creates monsters
// in the game.
//=========================================================

#include "extdll.h"
#include "util.h"
#include "cbase.h"
#include "monsters.h"
#include "weapons.h"
#include "saverestore.h"

// Monstermaker spawnflags
#define	SF_MONSTERMAKER_START_ON	1 // start active ( if has targetname )
#define	SF_MONSTERMAKER_CYCLIC		4 // drop one monster every time fired.
#define SF_MONSTERMAKER_MONSTERCLIP	8 // Children are blocked by monsterclip

//=========================================================
// MonsterMaker - this ent creates monsters during the game.
//=========================================================
class CMonsterMaker : public CBaseMonster
{
public:
	void Spawn( void );
	void Precache( void );
	void KeyValue( KeyValueData* pkvd);
	void EXPORT ToggleUse ( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	void EXPORT CyclicUse ( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value );
	void EXPORT MakerThink ( void );
	void DeathNotice ( entvars_t *pevChild );// monster maker children use this to tell the monster maker that they have died.
	void MakeMonster( void );

	virtual int		Save( CSave &save );
	virtual int		Restore( CRestore &restore );

	static	TYPEDESCRIPTION m_SaveData[];
	
	string_t m_iszMonsterClassname;// classname of the monster(s) that will be created.
	
	int	 m_cNumMonsters;// max number of monsters this ent can create

	
	int  m_cLiveChildren;// how many monsters made by this monster maker that are currently alive
	int	 m_iMaxLiveChildren;// max number of monsters that this maker may have out at one time.

	float m_flGround; // z coord of the ground under me, used to make sure no monsters are under the maker when it drops a new child

	BOOL m_fActive;
	BOOL m_fFadeChildren;// should we make the children fadeout?

        // monstermaker is now a tangible object.
        int TakeDamage( entvars_t *pevInflictor, entvars_t *pevAttacker, float flDamage, int bitsDamageType );
        void Killed( entvars_t *pevAttacker, int iGib );
        void FadeOut ( void  );
        CSprite *m_pBall;    //foom
};

LINK_ENTITY_TO_CLASS( monstermaker, CMonsterMaker );

TYPEDESCRIPTION	CMonsterMaker::m_SaveData[] = 
{
	DEFINE_FIELD( CMonsterMaker, m_iszMonsterClassname, FIELD_STRING ),
	DEFINE_FIELD( CMonsterMaker, m_cNumMonsters, FIELD_INTEGER ),
	DEFINE_FIELD( CMonsterMaker, m_cLiveChildren, FIELD_INTEGER ),
	DEFINE_FIELD( CMonsterMaker, m_flGround, FIELD_FLOAT ),
	DEFINE_FIELD( CMonsterMaker, m_iMaxLiveChildren, FIELD_INTEGER ),
	DEFINE_FIELD( CMonsterMaker, m_fActive, FIELD_BOOLEAN ),
	DEFINE_FIELD( CMonsterMaker, m_fFadeChildren, FIELD_BOOLEAN ),
};


IMPLEMENT_SAVERESTORE( CMonsterMaker, CBaseMonster );

void CMonsterMaker :: KeyValue( KeyValueData *pkvd )
{
	
	if ( FStrEq(pkvd->szKeyName, "monstercount") )
	{
		m_cNumMonsters = atoi(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if ( FStrEq(pkvd->szKeyName, "m_imaxlivechildren") )
	{
		m_iMaxLiveChildren = atoi(pkvd->szValue);
		pkvd->fHandled = TRUE;
	}
	else if ( FStrEq(pkvd->szKeyName, "monstertype") )
	{
		m_iszMonsterClassname = ALLOC_STRING( pkvd->szValue );
		pkvd->fHandled = TRUE;
	}
	else
		CBaseMonster::KeyValue( pkvd );
}


void CMonsterMaker :: Spawn( )
{

        Precache();
        SET_MODEL(ENT(pev), "models/spawner.mdl");
        pev->movetype = MOVETYPE_NONE;
        pev->solid  = SOLID_BBOX;

        //droptoground
        TraceResult tr;
        UTIL_TraceLine ( pev->origin, pev->origin - Vector ( 0, 0, 2048 ), ignore_monsters, ENT(pev), &tr );
        UTIL_SetOrigin( pev, tr.vecEndPos );

        UTIL_SetSize( pev, Vector(-32,-32,0), Vector(32,32,128) );

        pev->takedamage = DAMAGE_YES;
        pev->health = gSkillData.apacheHealth;
        pev->angles.y  = RANDOM_LONG(1,359);//random direction
        m_bloodColor            = DONT_BLEED;


        //SetActivity( ACT_IDLE );
        //pev->sequence = 0;
        pev->framerate = 8;


	m_cLiveChildren = 0;

	if ( !FStringNull ( pev->targetname ) )
	{
		if ( pev->spawnflags & SF_MONSTERMAKER_CYCLIC )
		{
			SetUse ( &CMonsterMaker::CyclicUse );// drop one monster each time we fire
		}
		else
		{
			SetUse ( &CMonsterMaker::ToggleUse );// so can be turned on/off
		}

		if ( FBitSet ( pev->spawnflags, SF_MONSTERMAKER_START_ON ) )
		{// start making monsters as soon as monstermaker spawns
			m_fActive = TRUE;
			SetThink ( &CMonsterMaker::MakerThink );
		}
		else
		{// wait to be activated.
			m_fActive = FALSE;
			SetThink ( &CMonsterMaker::SUB_DoNothing );
		}
	}
	else
	{// no targetname, just start.
			pev->nextthink = gpGlobals->time + m_flDelay;
			m_fActive = TRUE;
			SetThink ( &CMonsterMaker::MakerThink );
	}

	if ( m_cNumMonsters == 1 )
	{
		m_fFadeChildren = FALSE;
	}
	else
	{
		m_fFadeChildren = TRUE;
	}

	m_flGround = 0;
}

void CMonsterMaker :: Precache( void )
{
	CBaseMonster::Precache();

        PRECACHE_MODEL("models/spawner.mdl");
	UTIL_PrecacheOther( STRING( m_iszMonsterClassname ) );
}













void CMonsterMaker :: Killed( entvars_t *pevAttacker, int iGib )
{

pev->deadflag = DEAD_DYING;
CBaseMonster :: FCheckAITrigger();


        pev->takedamage = DAMAGE_NO;

        UTIL_ScreenShake( pev->origin, 20.0, 150.0, 2.0, 512.0 );

        //foom! g_sModelIndexFoom create my foom letters.
        m_pBall = CSprite::SpriteCreate( "models/foom01.mdl", pev->origin, FALSE );
        m_pBall->pev->velocity = Vector( 0, 0, RANDOM_LONG(512,768) );//upward float
        m_pBall->pev->angles.y = RANDOM_LONG(1,359);//random yaw
        m_pBall->pev->avelocity.y = RANDOM_LONG(-96,96);//random spinning.

        m_pBall->SetThink( &CMonsterMaker::SUB_Remove );
        m_pBall->pev->nextthink = gpGlobals->time + 2;//lifespan



        switch (RANDOM_LONG(1,6)) //drop a powerup.
        {
                case 1: DropItem( "weapon_mp5", pev->origin, Vector(0,RANDOM_LONG(1,350),0) );    break;
                case 2: DropItem( "weapon_shotgun", pev->origin, Vector(0,RANDOM_LONG(1,350),0) );break;
                case 3: DropItem( "weapon_crossbow", pev->origin, Vector(0,RANDOM_LONG(1,350),0) );break;
                case 4: DropItem( "weapon_rpg", pev->origin, Vector(0,RANDOM_LONG(1,350),0) );     break;
                case 5: DropItem( "weapon_handgrenade", pev->origin, Vector(0,RANDOM_LONG(1,350),0) );break;
                case 6: DropItem( "weapon_hornetgun", pev->origin, Vector(0,RANDOM_LONG(1,350),0) );  break;
        }




/*bigexplosion.
        MESSAGE_BEGIN( MSG_PVS, SVC_TEMPENTITY, pev->origin );
                WRITE_BYTE( TE_SPRITE );
                WRITE_COORD( pev->origin.x );     // pos
                WRITE_COORD( pev->origin.y ); 
                WRITE_COORD( pev->origin.z ); 
                WRITE_SHORT( g_sModelIndexStar );              // model
                WRITE_BYTE( 24 );                                // size * 10
                WRITE_BYTE( 255 );                      // brightness
        MESSAGE_END();
  */

/*
		MESSAGE_BEGIN( MSG_PVS, SVC_TEMPENTITY, pev->origin );
			WRITE_BYTE( TE_BEAMCYLINDER );
			WRITE_COORD( pev->origin.x);
			WRITE_COORD( pev->origin.y);
			WRITE_COORD( pev->origin.z);
			WRITE_COORD( pev->origin.x);
			WRITE_COORD( pev->origin.y);
                        WRITE_COORD( pev->origin.z + 512 ); // reach damage radius over .2 seconds
                        WRITE_SHORT( g_sModelIndexWExplosion );
			WRITE_BYTE( 0 ); // startframe
			WRITE_BYTE( 0 ); // framerate
                        WRITE_BYTE( 6 ); // life
			WRITE_BYTE( 32 );  // width
			WRITE_BYTE( 0 );   // noise
                        WRITE_BYTE( 255 );   // r, g, b
                        WRITE_BYTE( 255 );   // r, g, b
                        WRITE_BYTE( 255 );   // r, g, b
                        WRITE_BYTE( 255 ); // brightness
                        WRITE_BYTE( 0 );                // speed
		MESSAGE_END();
  */
        CBaseMonster :: Killed( pevAttacker, GIB_NEVER );

}

int CMonsterMaker :: TakeDamage( entvars_t *pevInflictor, entvars_t *pevAttacker, float flDamage, int bitsDamageType )
{

        //ALERT(at_console,"MM HEALTH: %f\n",pev->health - flDamage);

        if      ((pev->health < (gSkillData.apacheHealth * 0.25))    )
        {
                if (pev->skin != 3)
                pev->skin = 3;
        }
        else if ((pev->health < (gSkillData.apacheHealth * 0.5))     )
        {
                if (pev->skin != 2)
                pev->skin = 2;
        }
        else if ((pev->health < (gSkillData.apacheHealth * 0.75))     )
        {
                if (pev->skin != 1)
                pev->skin = 1;
        }


	return CBaseMonster::TakeDamage( pevInflictor, pevAttacker, flDamage, bitsDamageType );
}





//=========================================================
// MakeMonster-  this is the code that drops the monster
//=========================================================
void CMonsterMaker::MakeMonster( void )
{
	edict_t	*pent;
	entvars_t		*pevCreate;

	if ( m_iMaxLiveChildren > 0 && m_cLiveChildren >= m_iMaxLiveChildren )
	{// not allowed to make a new one yet. Too many live ones out right now.
		return;
	}

	if ( !m_flGround )
	{
		// set altitude. Now that I'm activated, any breakables, etc should be out from under me. 
		TraceResult tr;

		UTIL_TraceLine ( pev->origin, pev->origin - Vector ( 0, 0, 2048 ), ignore_monsters, ENT(pev), &tr );
		m_flGround = tr.vecEndPos.z;
	}

	Vector mins = pev->origin - Vector( 34, 34, 0 );
	Vector maxs = pev->origin + Vector( 34, 34, 0 );
	maxs.z = pev->origin.z;
	mins.z = m_flGround;

	CBaseEntity *pList[2];
	int count = UTIL_EntitiesInBox( pList, 2, mins, maxs, FL_CLIENT|FL_MONSTER );
	if ( count )
	{
		// don't build a stack of monsters!
		return;
	}

	pent = CREATE_NAMED_ENTITY( m_iszMonsterClassname );

	if ( FNullEnt( pent ) )
	{
		ALERT ( at_console, "NULL Ent in MonsterMaker!\n" );
		return;
	}
	
	// If I have a target, fire!
	if ( !FStringNull ( pev->target ) )
	{
		// delay already overloaded for this entity, so can't call SUB_UseTargets()
		FireTargets( STRING(pev->target), this, this, USE_TOGGLE, 0 );
	}

	pevCreate = VARS( pent );
        pevCreate->origin = pev->origin;
        pevCreate->angles = pev->angles;
        //bc

        int randx,randy;
        switch (RANDOM_LONG(0,3))
        {
        case 0: randx = (RANDOM_LONG(128,384));randy = (RANDOM_LONG(128,384));break;
        case 1: randx = (RANDOM_LONG(128,384));randy = (RANDOM_LONG(-128,-384));break;
        case 2: randx = (RANDOM_LONG(-128,-384));randy = (RANDOM_LONG(128,384));break;
        case 3: randx = (RANDOM_LONG(-128,-384));randy = (RANDOM_LONG(-128,-384));break;
        }

        pevCreate->velocity = Vector(randx,randy,RANDOM_LONG(128,384));
        SetBits( pevCreate->spawnflags, SF_MONSTER_FALL_TO_GROUND );

	// Children hit monsterclip brushes
	if ( pev->spawnflags & SF_MONSTERMAKER_MONSTERCLIP )
		SetBits( pevCreate->spawnflags, SF_MONSTER_HITMONSTERCLIP );

	DispatchSpawn( ENT( pevCreate ) );
	pevCreate->owner = edict();

	if ( !FStringNull( pev->netname ) )
	{
		// if I have a netname (overloaded), give the child monster that name as a targetname
		pevCreate->targetname = pev->netname;
	}

	m_cLiveChildren++;// count this monster
	m_cNumMonsters--;

	if ( m_cNumMonsters == 0 )
	{
		// Disable this forever.  Don't kill it because it still gets death notices
		SetThink( NULL );
		SetUse( NULL );
	}
}

//=========================================================
// CyclicUse - drops one monster from the monstermaker
// each time we call this.
//=========================================================
void CMonsterMaker::CyclicUse ( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	MakeMonster();
}

//=========================================================
// ToggleUse - activates/deactivates the monster maker
//=========================================================
void CMonsterMaker :: ToggleUse ( CBaseEntity *pActivator, CBaseEntity *pCaller, USE_TYPE useType, float value )
{
	if ( !ShouldToggle( useType, m_fActive ) )
		return;

	if ( m_fActive )
	{
		m_fActive = FALSE;
		SetThink ( NULL );
	}
	else
	{
		m_fActive = TRUE;
		SetThink ( &CMonsterMaker::MakerThink );
	}

	pev->nextthink = gpGlobals->time;
}

//=========================================================
// MakerThink - creates a new monster every so often
//=========================================================
void CMonsterMaker :: MakerThink ( void )
{
	pev->nextthink = gpGlobals->time + m_flDelay;

	MakeMonster();
}


//=========================================================
//=========================================================
void CMonsterMaker :: DeathNotice ( entvars_t *pevChild )
{
	// ok, we've gotten the deathnotice from our child, now clear out its owner if we don't want it to fade.
	m_cLiveChildren--;

	if ( !m_fFadeChildren )
	{
		pevChild->owner = NULL;
	}
}


